; Boot sector to load up an OS kernel
;
; Copyright (C) 1999 Thomas Novelli
;	Retro is distributed under the terms of the GNU LGPL, version 2 or
;	greater. (For the text of the license, see the file COPYING.LIB)
;
; Platform: Intel x86
;	Thanks for the tips & examples: Linus Torvalds, John S. Fine,
;	M. J. Vines, Francois-Rene Rideau
;
; To do:
;	Eventually, load kernel from a file/object (after the persistent
;	storage system has stabilized)
;
;
; Here's how it works:
;
; BIOS: 
;  1. Loads Sector 1, Track 0, Head 0 to address 7C00
;  2. Checks the last two bytes of the bootsector, which must contain
;     the signature 0xAA55, or it says "Operating system not found".
;  3. Loads DL with:
;	0x00 if booted from drive A
;	0x80 if booted from drive C
;	(Note: What about bootable CD's? 0x81, 0x82, etc?)
;  4. Jumps to 0000:7C00
;
; Boot loader:
;  1. Initializes segments & stack
;  2. Prints a message
;  3. Loads the rest of the boot loader, starting at address 7E00
;
; Initialization:
;  1. Checks some things, like the CPU type, RAM availble, etc.
;  2. Starts protected mode (386+ only)
;
; Kernel:
;  The kernel (core of the operating system) is responsible for loading
;  and managing all programs/objects.

; Memory map:
;	0000 to 03FF	; Real mode Interrupt Vector Table
;	0400 to 049F	; BIOS data area
;	7000 to 7BFF	; Stack for boot loader
;	7C00 to 7DFF	; Boot sector
;	7E00 to ????	; Extended boot loader & OS kernel

BITS 16

;Note: we could detect the boot address if that helps. CALL $+2 and POP..

%define BOOTSEG		0x7C0	; BIOS puts the boot sector at address 7C00
%define STACKSEG	0x700	; This area should be free for the stack...

;; Some debugging macros
%macro hexp 1	;; hexp <number>
	pusha
	push word %1
	mov	bp, sp
	call	hex_print
	call	print_nl
	popa
%endmacro

%define hang	jmp short $


	ORG	0x7C00

boot_start:
	jmp short start
	nop


;--------------------------------------------------------------------
;
; Data block
;
data:
.sects	db	num_sectors	;Number of extra sectors used by boot loader
;.sects	db	10

.moddir	dd	36		;Module directory (sector #)
.modord	dd	37		;Module loading order (sector #)
times	(0x3B - ($-data))	db	0


;--------------------------------------------------------------------


start:
	xor	ax, ax		; Initialize segments
	mov     ds, ax		;ds,es = 0
	mov	es, ax
	mov	ax, STACKSEG
	mov     ss, ax
	mov     sp, 0xC00

	mov	[boot_drive], dl	;BIOS passes this along; save it.

; An attempt to solve Tril's problem with his CGA. It flickers for a few
; seconds, then winds up in 40x25 mode somehow, with "Welcome to DDR.." at
; the upper left.
;
;	mov	ax, 0x0003		; Set video mode 03h
;	int	0x10

        mov     bp, boot_msg		; Startup message
        call    print_string

; Load kernel
	call	probe

	mov	cx, 0x07E0
	mov	es, cx
	mov	cl, [data.sects]	;# sectors
	xor	ch, ch
	call	read_sectors

        mov     dx,03F2h	;Shut off floppy motor
        xor     al,al
        out     dx,al

	jmp	init

init: 				; BOGUS, jmps to 512

;--------------------------------------------------------------------


sect	dw	1	;Current location (0-based)
track	dw	0
head	dw	0


read_sectors:
;;
;; CX = number of sectors to load
;; ES:0 = address to load at
;;
;; John Fine's BOOTP.ASM has a more elegant 'read_sectors'. I should steal
;; it :)
;;
.track
	mov	ax, [sectors]
	sub	ax, [sect]	;AX = # sectors to load on this track
	cmp	ax, cx		;ax = min (ax, cx)
	jb	.1
	mov	ax, cx

.1	mov	bx, es		;Crossing a 64k boundary? Lower AX.
	mov	dx, bx
	shr	dx, 12		;round DX up to nearest 64k
	shl	dx, 12
	add	dx, 0x1000
	sub	dx, bx		;DX = # sectors remaining in 64k segment
	shr	dx, 5
	cmp	ax, dx		;ax = min (ax, dx)
	jb	.2
	mov	ax, dx
.2

	call	read_track
	sub	cx, ax

	mov	dx, ax		;Increment ES
	shl	dx, 5
	mov	bx, es
	add	bx, dx
	mov	es, bx

	add	ax, [sect]	;Increment current sector
	cmp	ax, [sectors]
	jb	.t1
	xor	ax, ax		;Flip head
	inc word [head]
	cmp word [head], 2
	jb	.t1
	mov word [head], 0	;Next track
	inc word [track]
.t1
	mov	[sect], ax
	or	cx, cx
	jnz	.track
	ret


read_track:
;; AX = # of sectors
	pusha
	mov	cx, 3		;# of retries
.retry
	push	ax
	push	cx
	xor	bx, bx
	mov	cl, [sect]
	inc	cl		;1-based
	mov	ch, [track]
	mov	dl, [boot_drive]
	mov	dh, [head]
	mov	ah, 2
	int	0x13
	jc	.error
	add	sp, 2*2		;drop ax & cx
	popa
	ret
.error
	pop	cx
	pop	ax
	loop	.retry
	call	debug
	mov	bp, disk_error_msg	;Then give up.
	call	print_string
	jmp short $


;--------------------------------------------------------------------


probe:
;
; Get disk parameters (by trial & error). Try 36, 18, 15, and 9 sectors
;
        mov     si, DiskSizes
.loop
        lodsb   ;AL=[SI]
        cbw     ;AX=AL (extended to word length)
        mov     [sectors], ax
        cmp     si, DiskSizes+4
        jae     .done
        xchg    ax, cx
        xor     dx, dx
        xor     bl, bl
        mov     bh, [sectors]
        inc     bh
        shl     bh, 1
        mov     ax, 0x0201
        int     0x13
        jc      .loop
.done
	ret



debug:
	pusha
	push	dx
	push	cx
	push	bx
	push	ax
	pushf
	call    print_all
	add	sp, 5*2
	popa
	ret



print_all:
;
; Register debugging dump (borrowed from Linux bootloader)
; Expects this on stack: dx, cx, bx, ax, errorcode
;
        mov     cx,5    ;1 error code + 4 registers
        mov     bp,sp
.print_loop:
        push    cx      ;save CX reg
        call    print_nl

        cmp     cl,5    ;if (cx < 5) skip the next part
        jae     .no_reg

        mov     ax,0E05h + 'A' - 1
        sub     al,cl
        int     10h
        mov     al,'X'
        int     10h
        mov     al,':'
        int     10h
.no_reg:
        add     bp,2
        call    hex_print
        pop     cx
        loop    .print_loop
        ret



; Supporting print functions
print_nl:
        mov     ax,0E0Dh        ;Service 0E=TTY character output
        int     10h
        mov     al,0Ah
        int     10h
        ret



hex_print:
;
; Prints the word at (ss:bp)
;
        mov     cx,4
        mov     dx,[bp]
print_digit:
        rol     dx,4
        mov     ax,0E0Fh
        and     al,dl
        add     al,90h
        daa
        adc     al,40h
        daa
        int     10h
        loop    print_digit
        ret



print_string:
;
; Let ES:BP point to a null-terminated string
;
        mov     ah,0Eh
        mov     bx,0            ;Set video page & Text olor
.l1     cmp     [ds:bp],byte 0
        je      .end
        mov     al,[ds:bp]
        int     10h
        inc     bp
        jmp     .l1
.end    ret

;--------------------------------------------------------------------

boot_msg	db "RETRO #9 bootloader",13,10,0
;boot_msg	db "RETRO #9 bootloader, ",DATE," by ",USER,13,10,0

disk_error_msg	db 13,10,"Disk Error",13,10,0


;Disk specs
sectors:        dw      0
tracks:		dw	80
heads:		dw	2

DiskSizes	db 36,18,15,9

boot_drive	db 0		; Boot drive (initial value of DL)


times	510-($-$$)	db 0	; skip to end of sector
	dw	0xAA55		; boot sector signature




modules_start:
; Any modules concatenated to the boot image will be right here.
;
; Note: The boot sector won't load the modules unless the 'num_sectors'
; field in the data block is increased to account for the modules.  Do this
; by running the 'adjboot' utility after tacking on the modules.


;boot_size	equ	$-boot_start
;num_sectors	equ	(boot_size - (boot_size % 512) ) / 512
num_sectors	equ	1
